﻿

#include <iostream>
#include <vector>
#include <cmath>
#include <exception>
#include <GL/glut.h>
#include "othello.hpp"
#include "othelloBoard.hpp"


#define M_PI 3.1415926535897932384626433832795

using namespace std;

//二次元ベクトル
struct Vector2D {
	double X, Y;
};

//オセロ盤を表示する関数
class BoardRenderer {
private:
	vector<Vector2D> verticies;
	//円を表示する
	void RenderCircle() {
		glBegin(GL_POLYGON);
		for (int i = 0; i < verticies.size(); i++)
			glVertex2d(verticies[i].X, verticies[i].Y);
		glEnd();
	}
	//円の周りの線を表示する
	void RenderCircleFrame() {
		glBegin(GL_LINE_LOOP);
		for (int i = 0; i < verticies.size(); i++)
			glVertex2d(verticies[i].X, verticies[i].Y);
		glEnd();
	}
public:
	//コンストラクタ。円の表示に使うテーブル(verticies)を初期化		
	BoardRenderer() {
		int count = 100;
		for (int i = 0; i < count; i++) {			
			Vector2D a;
			double theta = (double)i * 2 * M_PI / count; 
			a.X = cos(theta);
			a.Y = sin(theta);
			verticies.push_back(a);
		}
	}
	//オセロ盤を表示する関数
	void Render(Board a, int turn) {		
		glClearColor(0, 0.5, 0, 1.0);
		glClear(GL_COLOR_BUFFER_BIT);
		//縦横の線を表示
		for (int i = 0; i <= 8; i++) {
			double x = 2 * ((double)i / 8 - 0.5);

			glColor3d(1, 1, 1);

			glBegin(GL_LINE_LOOP);
			glVertex2d(-1, x);
			glVertex2d(1, x);
			glEnd();

			glBegin(GL_LINE_LOOP);
			glVertex2d(x, -1);
			glVertex2d(x, 1);
			glEnd();
		}

		glMatrixMode(GL_MODELVIEW);		

		//オセロ盤を表示
		for (int i = 0; i < 8; i++)
			for (int j = 0; j < 8; j++) {
				int disk = a.GetDisk(i + 8 * j);
				if (disk == 2) {
					bool placable;
					if (turn == 0)
						placable = a.GetPlacable(i + 8 * j);
					else {
						a.Flip();
						placable = a.GetPlacable(i + 8 * j);
						a.Flip();
					}
					if (!placable)
						continue;
				}

				glPushMatrix();
				{
					double x = 2 * (((double)i + 0.5) / 8 - 0.5);
					double y = 2 * (((double)j + 0.5) / 8 - 0.5);
					glTranslated(x, y, 0);
					glScaled(0.8 / 8, 0.8 / 8, 0);
					if (disk == 0) //黒石
						glColor3d(0, 0, 0);
					else if (disk == 1) //白石
						glColor3d(1, 1, 1);
					else {
						//置ける場所を表示
						glScaled(0.4, 0.4, 0);
						glColor3d(0.8, 0.8, 0.4);
					}
					RenderCircle();
					glColor3d(0.5, 0.5, 0.5);
					RenderCircleFrame();

					//確定石マークを表示
					if (a.GetFixed(i + 8 * j)) {
						glScaled(0.4, 0.4, 0);
						glColor3d(0.8, 0.6, 0.7);
						RenderCircle();
					}
				}
				glPopMatrix();
			}		
	}
};

BoardRenderer *renderer; //オセロ盤を表示する
Othello *othello; //オセロゲーム自体を管理する
UserOthelloPlayer *user1; //人間のプレーヤー
OthelloPlayer1 *player1; //コンピュータのプレーヤー
int windowWidth = 1; //ウィンドウの幅
int windowHeight = 1; //ウィンドウの高さ

//マウスイベント
void mouse(int button, int state, int x, int y) {	
	if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) {
		//ゲームが終了していたら新しいゲームを始める
		if (othello->GetCurrent().Gameover()) {
			cout << player1->PrunnedRateLog() << endl;
			player1->Initialize();
			othello->NewGame();
		}
		else {
			//人間のプレーヤークラスにメッセージを送る
			x = 8 * x / windowWidth;
			y = 8 * y / windowHeight;
			user1->Message(x, y);
		}
	}
}

//画面のリサイズイベント
void reshape(int width, int height) {
	windowWidth = std::max(1, width);
	windowHeight = std::max(1, height);
	glViewport(0, 0, width, height);
}

//メインループ。必要に応じて再描画イベントを送る
void idle(void) {
	static Board prev;
	if (prev == othello->GetCurrent())
		return;	
	prev = othello->GetCurrent();
	glutPostRedisplay();
}

//描画関数
void display(void) {
	glClear(GL_COLOR_BUFFER_BIT);
	glBegin(GL_LINE_LOOP);
	glVertex2d(-0.9, -0.9);
	glVertex2d(0.9, -0.9);
	glVertex2d(0.9, 0.9);
	glVertex2d(-0.9, 0.9);
	glEnd();

	renderer->Render(othello->GetCurrent(), othello->GetTurn());

	glFlush();
}

//実験データを取るための関数。通常モードでは使われない
void experiment() {
	const int PlayerCount = 20;
	OthelloPlayer1 *players[PlayerCount];
	double scores[PlayerCount];
	for (int i = 0; i < PlayerCount; i++) 			
		scores[i] = 1;

	//プレイヤーの初期化
	for (int i = 0; i < PlayerCount; i++) {			
		players[i] = new OthelloPlayer1((double)(i / 2) / (PlayerCount / 2 - 1), i % 2 == 0);
	}

	//リーグ戦
	for (int i = 0; i < PlayerCount; i++) 
		for (int j = 0; j < PlayerCount; j++) {
			if (i == j)
				continue;
			players[i]->Initialize();
			players[j]->Initialize();
			othello = new Othello(players[i], players[j]);
			othello->WaitFor();
			int black = othello->GetCurrent().BlackCount();
			int white = othello->GetCurrent().WhiteCount();
			int point;
			if (black > white)
				point = 1;
			else if (black < white)
				point = -1;
			else
				point = 0;
			scores[i] += point;
			scores[j] -= point;
		}
	for (int i = 0; i < PlayerCount; i++) {
		scores[i] /= 2 * (PlayerCount - 1);
		scores[i] = (1 + scores[i]) / 2;
	}
			
	//結果表示
	cout << "parameter1, 勝率, log(カット率), log(カット率)の分散" << endl;
	cout << "確定石の数使用" << endl;
	for (int i = 0; i < PlayerCount; i++) {		
		if (players[i]->UseFixed())
			cout << players[i]->Parameter1() << ", " << scores[i] << ", " << players[i]->PrunnedRateLog() << ", " 
				<< players[i]->PrunnedRateVariLog() << endl;
	}
	cout << "通常の石の数使用" << endl;
	for (int i = 0; i < PlayerCount; i++) {		
		if (!players[i]->UseFixed())
			cout << players[i]->Parameter1() << ", " << scores[i] << ", " << players[i]->PrunnedRateLog() << ", " 
				<< players[i]->PrunnedRateVariLog() << endl;
	}
}

void experiment2() {
	OthelloPlayer1 *players[2];

	//プレイヤーの初期化
	for (int i = 0; i < 2; i++) {			
		players[i] = new OthelloPlayer1(0.3, true);
	}
	othello = new Othello(players[0], players[1]);
	othello->WaitFor();
			
	//結果表示
	cout << "枝狩り率" << players[0]->PrunnedRateLog() << ", " << players[0]->PrunnedRateVariLog() << endl;
}

//エントリーポイント
int main(int argc, char *argv[]) {

	if (false) {
		//評価関数の実験モード
		experiment();

		int a;
		cin >> a;
	}
	else if (false) {
		//アルファベータカットの実験モード
		experiment2();

		int a;
		cin >> a;
	}
	else {
		//通常モード
	
		//プレーヤーとオセロゲームの初期化
		user1 = new UserOthelloPlayer();
		player1 = new OthelloPlayer1(0.3, true); //実験結果を見て決めた
		othello = new Othello(user1, player1);

		//オセロ盤を表示するクラスの初期化
		renderer = new BoardRenderer();
	
		//OpenGLの初期化
		glutInit(&argc, argv);
		glutCreateWindow(argv[0]);
		glutDisplayFunc(display);
		glutMouseFunc(mouse);
		glutReshapeFunc(reshape);
		glutIdleFunc(idle);

		glMatrixMode(GL_PROJECTION);
		glScaled(1, -1, 1);

		glutMainLoop();
	}

	return 0;
}